//+------------------------------------------------------------------+
//|                                                          xlv.mq5 |
//|                                  Copyright 2022, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2022, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#property version   "1.00"
#include                            <Cgini.mqh>
#include                            <Trade\Trade.mqh>
#include                            <Math\Stat\stat.mqh>
#include                   		   <Trade\SymbolInfo.mqh>
#include                            <Indicators\Indicators.mqh>
#include                            <Math\Alglib\dataanalysis.mqh>
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CTrade                              __T;
Cgini _g;
CMLPBase _base;
CMLPTrain _train;
CMatrixDouble _xy;
CMLPReport _report;
CMLPTrainer _trainer;
CMultilayerPerceptron _mlp;
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum Edataframe
  {
      DATA_QUARTER=65,
      DATA_MONTH=21,
      DATA_WEEK=5,
      DATA_DAY=1,
  };
//+------------------------------------------------------------------+
//| INPUTS                                                           |
//+------------------------------------------------------------------+
#define              __WEEK PeriodSeconds(PERIOD_W1)
#define              __TIMER PeriodSeconds(_Period)
#define              __QUARTER 13*(__WEEK/__TIMER)
#define              __MARKET_OPEN 15
#define              __MARKET_CLOSE 20
input string         __data_set="compound_set";
sinput Edataframe    __data_frame=DATA_QUARTER;
input int            __hidden=1;
input int            __hidden_1_size=5;
input int            __hidden_2_size=3;
input int            __training_points=8;
sinput int           __training_restarts=3;
static int           __DATA_INDEX;
static bool          __IN_TRAINING;
MqlDateTime          __DATETIME;
//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
int OnInit()
  {
   if(Period()>=PERIOD_D1)
   {
      printf(__FUNCSIG__+" period should be below: "+EnumToString(PERIOD_D1));
      return(INIT_PARAMETERS_INCORRECT);
   }
   //
   if(_g.Get(__data_set))
   {
      _g.Set(__data_set);
      
      if(__training_points>=_g.xy.Rows()-1)
      {
         printf(__FUNCSIG__+" too many training points given sample size. ");
         return(INIT_PARAMETERS_INCORRECT);
      }
      
      int _inputs=_g.xy.Cols()-1;
      int _outputs=1;
      
      printf(__FUNCSIG__+" model x independents, and y classifiers are: "+IntegerToString(_inputs)+", and: "+IntegerToString(_outputs));
         
      _xy.Resize(__training_points,_inputs+_outputs);
      
      if(!ReadWeights())
      {
         _mlp.m_weights.Fill(1.0);
      }
      
      //
      
      if(__hidden==0)
      {
         _base.MLPCreate0(_inputs,_outputs,_mlp);
      }
      else if(__hidden==1)
      {
         if(__hidden_1_size<=_inputs)
         {
            printf(__FUNCSIG__+" hidden 1 size should exceed inputs. ");
            return(INIT_PARAMETERS_INCORRECT);
         }
         _base.MLPCreate1(_inputs,__hidden_1_size,_outputs,_mlp);
      }
      else if(__hidden==2)
      {
         if(__hidden_1_size<=_inputs)
         {
            printf(__FUNCSIG__+" hidden 1 size should exceed inputs. ");
            return(INIT_PARAMETERS_INCORRECT);
         }
         if(__hidden_2_size<=_outputs)
         {
            printf(__FUNCSIG__+" hidden 2 size should exceed outputs. ");
            return(INIT_PARAMETERS_INCORRECT);
         }
         _base.MLPCreate2(_inputs,__hidden_1_size,__hidden_2_size,_outputs,_mlp);
      }
      else if(__hidden>2||__hidden<0)
      {
         printf(__FUNCSIG__+" invalid number of hidden layers should be 0, 1, or 2. ");
         return(INIT_PARAMETERS_INCORRECT);
      }
      
      _train.MLPCreateTrainer(_inputs,_outputs,_trainer);
      
      __IN_TRAINING=false;
   }
   else
   {
      printf(__FUNCSIG__+" missing csv file... ");
      return(INIT_PARAMETERS_INCORRECT);
   }
   
   
//--- create timer
   EventSetTimer(__TIMER);
   
//---
   return(INIT_SUCCEEDED);
  }
//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
//--- destroy timer
   EventKillTimer();
   
  }
//+------------------------------------------------------------------+
//| Read Weights                                                     |
//+------------------------------------------------------------------+
bool ReadWeights()
  {
      CRowDouble _p;
      double _p_read[];
      ResetLastError();
      int _read_handle=FileOpen("_SPDR\\MLP_P_"+__data_set+"_"+IntegerToString(__hidden)+"_"+IntegerToString(__hidden_1_size)+"_"+IntegerToString(__hidden_2_size)+"_"+IntegerToString(__training_points)+".bin",FILE_READ|FILE_BIN|FILE_COMMON);
      if(_read_handle!=INVALID_HANDLE)
      {
         FileReadArray(_read_handle,_p_read);
         _p.Resize(ArraySize(_p_read));
         for(int i=0;i<_p.Size();i++){
            _p.Set(i,_p_read[i]);}
         
         printf(__FUNCSIG__+" read data = "+IntegerToString(_p.Size()));
         
         FileClose(_read_handle);
         
         _base.MLPImportTunableParameters(_mlp,_p);
         
         return(true);
      }
      else
      {
         printf(__FUNCSIG__+" File read failed, error "+IntegerToString(GetLastError()));
      }
      
      return(false);
  }
//+------------------------------------------------------------------+
//| Write Weights                                                    |
//+------------------------------------------------------------------+
bool WriteWeights(CRowDouble &Export)
  {
      double _p_write[];
      //--- open the file
      ResetLastError();
      int _write_handle=FileOpen("_SPDR\\MLP_P_"+__data_set+"_"+IntegerToString(__hidden)+"_"+IntegerToString(__hidden_1_size)+"_"+IntegerToString(__hidden_2_size)+"_"+IntegerToString(__training_points)+".bin",FILE_WRITE|FILE_BIN|FILE_COMMON);
      if(_write_handle!=INVALID_HANDLE)
      {
         ArrayResize(_p_write,Export.Size());
         Export.ToArray(_p_write);
         
         FileWriteArray(_write_handle,_p_write);
         
         printf(__FUNCSIG__+" written data = "+IntegerToString(Export.Size()));
         
         FileClose(_write_handle);
         
         return(true);
      }
      else
      {
         printf(__FUNCSIG__+" File write failed, error "+IntegerToString(GetLastError()));
      }
      
      return(false);
  }
//+------------------------------------------------------------------+
//| Expert tick function                                             |
//+------------------------------------------------------------------+
void OnTick()
  {
//---
   
  }
//+------------------------------------------------------------------+
//| Timer function                                                   |
//+------------------------------------------------------------------+
void OnTimer()
  {
//---
      datetime _now=TimeCurrent();
      //
      TimeToStruct(_now,__DATETIME);
      //
      if
      (
      __DATETIME.hour>=__MARKET_OPEN
      &&
      __DATETIME.hour<=__MARKET_CLOSE
      &&
      __DATETIME.day_of_week>=1
      &&
      __DATETIME.day_of_week<=5
      )
      {
         int _info=0;
         
         int _inputs=_g.xy.Cols()-1;
         int _outputs=1;
            
         if(IsNewDataRow(_now))
         {
            int _new_index=-1;
             
            for(int i=0;i<int(_g.w.Size());i++)
            {
               if(datetime(_g.w[i])<=_now && _now-_g.w[i]<=2*PeriodSeconds(PERIOD_W1))
               {
                  _new_index=i;break;
               }
            }
            //
            
            if(_new_index!=__DATA_INDEX && _new_index!=-1)
            {
               __DATA_INDEX=_new_index;
            }
            else
            {
               return;
            }
            //
            if(__DATA_INDEX<_xy.Rows())
            {
               return;
            }
            else if(__DATA_INDEX>=_xy.Rows())
            {
               for(int n=0;n<__training_points;n++)
               {
                  for(int i=0;i<_inputs;i++)
                  {
                     _xy.Set(n,i,_g.xy[__DATA_INDEX-n][i]);
                  }
                  //
                  //_xy.Set(n,_inputs,_g.xy[__DATA_INDEX-n][_inputs]);
                  double _new=iClose(Symbol(),Period(),n*__QUARTER);
                  double _old=iClose(Symbol(),Period(),(n+1)*__QUARTER);
                  _xy.Set(n,_inputs,_new-_old);
               } 
               
               _train.MLPSetDataset(_trainer,_xy,__training_points);
               //
               if(!__IN_TRAINING)
               {
                  _train.MLPStartTraining(_trainer,_mlp,false);
                  __IN_TRAINING=true;
               }
               else if(__IN_TRAINING)
               {
                  _train.MLPContinueTraining(_trainer,_mlp);
               }
               //
               double _x_inputs[],_y_outputs[];
               ArrayResize(_x_inputs,_inputs);
               ArrayResize(_y_outputs,_outputs);
               
               ArrayInitialize(_x_inputs,0.0);
               ArrayInitialize(_y_outputs,0.0);
               
               for(int i=0;i<_inputs;i++)
               {
                  _x_inputs[i]=_g.xy[__DATA_INDEX][i];
               }
            
               //forward feed...
               _base.MLPProcess(_mlp,_x_inputs,_y_outputs);
               printf(__FUNCSIG__+" y is: "+DoubleToString(_y_outputs[0])+", on index: "+IntegerToString(__DATA_INDEX));
               
               if(_y_outputs[0]>0.0)
               {
                  if(Longs()==0)
                  {
                     __T.Buy(0.1);
                  }
               }
               else if(_y_outputs[0]<0.0)
               {
                  if(Shorts()==0)
                  {
                     __T.Sell(0.1);
                  }
               }
            }
         }
      }
  }
//+------------------------------------------------------------------+
//| DataRow function                                                 |
//+------------------------------------------------------------------+
bool IsNewDataRow(datetime Time)
   {
      if(__data_frame==DATA_QUARTER)
      {
         return(IsNewQuarter(Time));
      }
      else if(__data_frame==DATA_MONTH)
      {
         return(IsNewMonth(Time));
      }
      else if(__data_frame==DATA_WEEK)
      {
         return(IsNewWeek(Time));
      }
      else if(__data_frame==DATA_DAY)
      {
         return(true);
      }
      
      return(false);
   }
//+------------------------------------------------------------------+
//| Quarter function                                                 |
//+------------------------------------------------------------------+
bool IsNewQuarter(datetime Time)
  {
//---
      MqlDateTime _d;
      TimeToStruct(Time,_d);
      string _1=IntegerToString(_d.year)+".03.31 00:00";
      string _2=IntegerToString(_d.year)+".06.30 00:00";
      string _3=IntegerToString(_d.year)+".09.30 00:00";
      string _4=IntegerToString(_d.year-1)+".12.31 00:00";
      datetime _1st=StringToTime(_1);
      datetime _2nd=StringToTime(_2);
      datetime _3rd=StringToTime(_3);
      datetime _4th=StringToTime(_4);
      
      return
      (
      (uint(Time)-uint(_1st)<uint(__WEEK)+PeriodSeconds(PERIOD_H6) && Time>_1st && _d.mon==4)
      ||
      (uint(Time)-uint(_2nd)<uint(__WEEK)+PeriodSeconds(PERIOD_H6) && Time>_2nd && _d.mon==7)
      ||
      (uint(Time)-uint(_3rd)<uint(__WEEK)+PeriodSeconds(PERIOD_H6) && Time>_3rd && _d.mon==10)
      ||
      (uint(Time)-uint(_4th)<uint(__WEEK)+PeriodSeconds(PERIOD_D1)+PeriodSeconds(PERIOD_D1)+PeriodSeconds(PERIOD_H6) && Time>_4th && _d.mon==1)
      );
  }
//+------------------------------------------------------------------+
//| Month function                                                   |
//+------------------------------------------------------------------+
bool IsNewMonth(datetime Time)
  {
//---
      MqlDateTime _d,_d_o;
      datetime _o=datetime(uint(Time)-uint(PeriodSeconds(PERIOD_D1)));
      TimeToStruct(Time,_d);TimeToStruct(_o,_d_o);
      
      return
      (
      (_d.day==1)
      ||
      (_d.day==2 && _d_o.day_of_week==0)
      ||
      (_d.day==3 && _d_o.day_of_week==0)
      );
  }
//+------------------------------------------------------------------+
//| Week function                                                    |
//+------------------------------------------------------------------+
bool IsNewWeek(datetime Time)
  {
//---
      MqlDateTime _d;
      TimeToStruct(Time,_d);
      
      return
      (
      (_d.day_of_week==1)
      );
  }
//+------------------------------------------------------------------+
//| Check for No. of Longs & Reverse Shorts.                         |
//+------------------------------------------------------------------+
int Longs()
   {
      int _longs=0;
      //
      int _p=PositionsTotal();
      //
      for(int i=_p-1;i>=0;i--)
      {
         if(PositionSelectByTicket(PositionGetTicket(i)) && PositionGetString(POSITION_SYMBOL)==Symbol())
         {
            if(PositionGetInteger(POSITION_TYPE)==POSITION_TYPE_SELL)
            {
               int _tries=0;
               do
               {
                  Sleep(1000*60);
                  _tries++;
               }
               while(!__T.PositionClose(PositionGetTicket(i)) && _tries<24*60);
            }
            else if(PositionGetInteger(POSITION_TYPE)==POSITION_TYPE_BUY)
            {
               _longs++;
            }
         }
      }
      
      return(_longs);
   }
//+------------------------------------------------------------------+
//| Check for No. of Shorts & Reverse Longs.                         |
//+------------------------------------------------------------------+
int Shorts()
   {
      int _shorts=0;
      //
      int _p=PositionsTotal();
      //
      for(int i=_p-1;i>=0;i--)
      {
         if(PositionSelectByTicket(PositionGetTicket(i)) && PositionGetString(POSITION_SYMBOL)==Symbol())
         {
            if(PositionGetInteger(POSITION_TYPE)==POSITION_TYPE_BUY)
            {
               int _tries=0;
               do
               {
                  Sleep(1000*60);
                  _tries++;
               }
               while(!__T.PositionClose(PositionGetTicket(i)) && _tries<24*60);
            }
            else if(PositionGetInteger(POSITION_TYPE)==POSITION_TYPE_SELL)
            {
               _shorts++;
            }
         }
      }
      
      return(_shorts);
   }
//+------------------------------------------------------------------+
//| Tester function                                                  |
//+------------------------------------------------------------------+
double OnTester()
  {
      double _result=TesterStatistics(STAT_PROFIT),_criteria=DBL_MIN;
      
      ResetLastError();
      int _read_handle=FileOpen("_SPDR\\MLP_criteria_"+__data_set+"_"+IntegerToString(__hidden)+"_"+IntegerToString(__hidden_1_size)+"_"+IntegerToString(__hidden_2_size)+"_"+IntegerToString(__training_points)+".bin",FILE_READ|FILE_BIN|FILE_COMMON);
      if(_read_handle!=INVALID_HANDLE)
      {
         _criteria=FileReadDouble(_read_handle);
         FileClose(_read_handle);
         
      }
      else
      {
         printf(__FUNCSIG__+" File read failed, error "+IntegerToString(GetLastError()));
         FileClose(_read_handle);
      }
      
      //
      printf(__FUNCSIG__+" criteria: "+DoubleToString(_criteria)+", result: "+DoubleToString(_result));
      
      if(_criteria<_result)
      {
         int _p_count;
         CRowDouble _p_export;
         _base.MLPExportTunableParameters(_mlp,_p_export,_p_count);
         WriteWeights(_p_export);
   
         ResetLastError();
         int _write_handle=FileOpen("_SPDR\\MLP_criteria_"+__data_set+"_"+IntegerToString(__hidden)+"_"+IntegerToString(__hidden_1_size)+"_"+IntegerToString(__hidden_2_size)+"_"+IntegerToString(__training_points)+".bin",FILE_WRITE|FILE_BIN|FILE_COMMON);
         if(_write_handle!=INVALID_HANDLE)
         {
            FileWriteDouble(_write_handle,_result);
            FileClose(_write_handle);
         }
         else
         {
            printf(__FUNCSIG__+" File write failed, error "+IntegerToString(GetLastError()));
            FileClose(_read_handle);
         }
      }
         
      return(_result);
  }
//+------------------------------------------------------------------+
