//+------------------------------------------------------------------+
//|                                                           nn.mq5 |
//|                                  Copyright 2022, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2022, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#property version   "1.00"
#property                           script_show_inputs
#include                            <Math\Alglib\dataanalysis.mqh>
#include                            <Cgini.mqh>
//+------------------------------------------------------------------+
//| INPUTS                                                           |
//+------------------------------------------------------------------+
input string   __file="compound_set";
input int      __max_hidden=5;
input double   __training_pct=0.50;
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
//---
      Cgini _g;
      if(_g.Get(__file))
      {
         _g.Set(__file);
      }
      return;
      //train _g.xy matrix through MLP model...
      
      int _inputs=_g.xy.Cols()-1;
      int _training_npoints=int(ceil(round(__training_pct*_g.xy.Rows())));
      int _outputs=1;
      
      CMLPBase _base;
      CMLPTrain _train;
      CMultilayerPerceptron _mlp;
      
      CMatrixDouble _xy;_xy.Resize(_training_npoints,_inputs+_outputs);
      
      int _info=0;
      CMLPReport _report;
      
      for(int h=1;h<=__max_hidden;h++)
      {
         _base.MLPCreate1(_inputs,h,_outputs,_mlp);
         
         for(int n=0;n<_training_npoints;n++)
         {
            for(int i=0;i<_inputs;i++)
            {
               _xy.Set(n,i,_g.xy[n][i+1]);
            }
         } 
         
         //train
         _train.MLPTrainLM(_mlp,_xy,_training_npoints,0.001,2,_info,_report);
         
         double _x_inputs[],_y_inputs[];
         ArrayResize(_x_inputs,_inputs);
         ArrayResize(_y_inputs,_outputs);
         
         static int _wins=0,_losses=0;
         static double _win=0.0,_loss=0.0;
         
         for(int n=_training_npoints;n<_g.xy.Rows()-_training_npoints;n++)
         {
            ArrayInitialize(_x_inputs,0.0);
            ArrayInitialize(_y_inputs,0.0);
            
            for(int i=0;i<_inputs-1;i++)
            {
               _x_inputs[i]=_g.xy[n][i+1];
            }
         
            //forward feed...
            _base.MLPProcess(_mlp,_x_inputs,_y_inputs);
            
            if((_y_inputs[0]>0.0 && _xy[n][_inputs-1]>0.0)||(_y_inputs[0]<0.0 && _xy[n][_inputs-1]<0.0))
            {
               _wins++;
               _win+=fabs(_xy[n][_inputs-1]);
            }
            else if((_y_inputs[0]<0.0 && _xy[n][_inputs-1]>0.0)||(_y_inputs[0]>0.0 && _xy[n][_inputs-1]<0.0))
            {
               _losses++;
               _loss+=fabs(_xy[n][_inputs-1]);
            }
         }
         //
         printf(__FUNCSIG__+" with: "+IntegerToString(h)+" hidden layers,  losses are: "+IntegerToString(_losses)+
                           ", and wins are:"+IntegerToString(_wins)+
                           ", and loss is:"+DoubleToString(_loss)+
                           ", with win at:"+DoubleToString(_win));
      }
  }
//+------------------------------------------------------------------+
